#include <iomanip>
#include <ranges>
#include <bcm2835.h>

#include "eeprom.hpp"
#include "util.hpp"

/* ADDRESS SETUP TIME
 * depends on clock frequency, see datasheet.  max 150@2MHz
 * address is set after setup time (tADS) and is being hold for address hold time (tAH)
 * 
 * 
 */
constexpr uint64_t addressSetupTimeUS = 1;  // 300e-9 s


EEPROM::EEPROM(const IOPins& ioPins, const AddressPins& addressPins, const Pin& pinCEb, const Pin& pinWEb, const Pin& pinOEb) 
    : ioPins(ioPins), ioMask(getMask(ioPins)), addressPins(addressPins), pinCEb(pinCEb), pinWEb(pinWEb), pinOEb(pinOEb) {
    // set to input and pullup
    auto setPinsInput = []<std::ranges::forward_range T>(const T& t) {
        for (auto it = t.begin(); it != t.end(); it++) {
            bcm2835_gpio_fsel(*it, BCM2835_GPIO_FSEL_INPT);
            /* bcm2835_gpio_set_pud(*it,  BCM2835_GPIO_PUD_UP); */
            bcm2835_gpio_set_pud(*it,  BCM2835_GPIO_PUD_OFF);
        }
    };
    setPinsInput(this->addressPins);
    setPinsInput(this->ioPins);
    setPinsInput(std::array<Pin, 3>{this->pinCEb, this->pinWEb, this->pinOEb});
}
EEPROM::~EEPROM() {
    setIODirection(INPUT);
}


void EEPROM::setIODirection(uint8_t direction) {
    if (IODirection == direction) { return; }
    for (size_t i = 0; i < ioPins.size(); i++) {
        bcm2835_gpio_fsel(ioPins.at(i), direction);
    }
}


void EEPROM::simulate(std::stop_token token) {
    std::println("Begin EEPROM simulation");
    
    using State = uint8_t;
    auto getState = [this](){ 
        State state = 0;
        state |= !readPin(pinCEb) << 1;
        state |= !readPin(pinWEb) << 2;
        state |= !readPin(pinOEb) << 3;
        return state;
    };
    auto isChipEnabled = [](State state) { return (state & (1 << 1)) != 0; };
    auto isWriteEnabled = [](State state) { return (state & (1 << 2)) != 0; };
    auto isOutputEnabled = [](State state) { return (state & (1 << 3)) != 0; };

    Address lastAddress = 0;
    State lastState = 0xff;

    while (!token.stop_requested()) {
        uint8_t state = getState();
        if (isChipEnabled(state)) {
            Address address = readAddress();
            if (isOutputEnabled(state) && (address != lastAddress || !isOutputEnabled(lastState))) { 
                setIODirection(OUTPUT);
                uint8_t data = rom.at(address);
                uint32_t valueMask = byteToMask(data, ioPins);
                bcm2835_gpio_write_mask(valueMask, ioMask);
                std::println("[0x{:4x} 0b{:16b}] -> {{0x{:2x} 0b{:8b}}}", address, address, data, data);
            }
            else if (isWriteEnabled(state) && (address != lastAddress || !isWriteEnabled(lastState))) {
                setIODirection(INPUT);
                uint8_t data = readData();
                rom.at(address) = data;
                std::println("[0x{:4x} 0b{:16b}] <- {{0x{:2x} 0b{:8b}}}", address, address, data, data);
            }
            lastAddress = address;
        }
        if (!isChipEnabled(state) && !(isOutputEnabled(state) || isWriteEnabled(state))) {
            setIODirection(INPUT);
        }
    }
    std::println("End EEPROM simulation");
}
